struct objectClass {
    oid	objid[32];
    int objid_length;
    int type;
    char description[128];
};

#define CLASS_KBOX	1
#define CLASS_CMU386	2
#define CLASS_CISCO	3
#define CLASS_CMUUNIX	4
#define CLASS_LANTERN	5
#define CLASS_NATBRIDGE	6
#define CLASS_SYNOPTICS1000 7
#define CLASS_SYNOPTICS3000 8
#define CLASS_ANNEX	9
#define CLASS_CMUBRIDGE	10
#define CLASS_CMUROUTER	11
#define CLASS_MULTINET	12
#define CLASS_ISODE	13
#define CLASS_PROTEON	14
#define CLASS_NYSERNET	15
#define CLASS_CMUKIP	16
#define CLASS_AFSSERVER 17

struct objectClass *objectClassByName();
struct objectClass *objectClassByType();
